#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-08-17
# @Filename: coordinate.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import warnings
from typing import Type, TypeVar

import numpy

from .defaults import VALID_WAVELENGTHS, WAVELENGTH
from .exceptions import CoordinateError, CoordIOError, CoordIOUserWarning
from .site import Site


T = TypeVar('T', bound="Coordinate", covariant=True)


def verifySite(kwargs, strict=True):
    """Search through kwargs for site parameter, and verify
    that it looks ok.  Otherwise raise a CoordIOError

    Parameters
    ------------
    kwargs : dict
        kwargs passed to Coordinate __new__
    strict : bool
        if True, a time must be specified on the site

    """
    if kwargs.get('site', None) is None:
        raise CoordIOError('Site must be passed')

    else:
        site = kwargs.get('site')
        if not isinstance(site, Site):
            raise CoordIOError('Site must be passed')
        if strict and site.time is None:
            raise CoordIOError(
                "Time of observation must be specified on Site"
            )


def verifyWavelength(kwargs, lenArray, strict=True):
    """Search through kwargs for wavelength parameter.  If not existent,
    return insert GFA wavelength to kwargs (of the right size).  If strict and
    wavelengths do not correspond to Apogee, Boss, or GFA wavelengths, raise a
    CoordIOError

    Parameters
    -----------
    kwargs : dict
        kwargs passed to Coordinate __new__
    lenArray : int
        length of the array to create if wavelength not present in kwargs
    strict : bool
        if strict is True, wavelengths may only be those corresponding to
        Apogee, Boss, or GFA

    """

    wls = kwargs.get("wavelength", None)
    if wls is None:
        # create an array of correct length default to gfa wavelength
        wls = numpy.zeros(lenArray) + WAVELENGTH
        warnings.warn(
            "Warning! Wavelengths not supplied, defaulting to %i angstrom" % WAVELENGTH,
            CoordIOUserWarning
        )
    else:
        if hasattr(wls, "__len__"):
            # array passed
            wls = numpy.array(wls, dtype=numpy.float64)
        else:
            # single value passed
            wls = numpy.zeros(lenArray) + float(wls)
        wlSet = set(wls)
        if strict and not wlSet.issubset(VALID_WAVELENGTHS):
            raise CoordIOError(
                "Invalid wavelength passed to FocalPlane \
                valid wavelengths are %s" % (str(VALID_WAVELENGTHS))
            )

    # modify
    kwargs["wavelength"] = wls


class Coordinate(numpy.ndarray):
    r"""Base class to represent a coordinate system.

    `.Coordinate` is a wrapper around the `numpy.ndarray` to represent a
    series of points in a given coordinate system. It handles most of the
    particularities of `subclassing a Numpy array
    <https://numpy.org/doc/stable/user/basics.subclassing.html>`__.

    A `.Coordinate` instance must be a 2D array of dimensions :math:`N\times M`
    in which N is the number of coordinate points, and M are their axis values.
    Extra parameters and arrays can be added to the coordinate object. For
    example, the following code minimally implements an ICRS coordinate system

    ::

        class ICRS(Coordinate):

        __extra_params__ ['name']
        __extra_arrays__ = ['epoch', 'pmra', 'pmdec', 'parallax', 'rvel']

    Now we can instantiate a set of coordinates ::

        >>> icrs = ICRS([[100., 20.], [200., 10], [300., 5.]],
                        pmra=[1, 2, 0], pmdec=[-0.5, 0.5, 0],
                        name='my_coordinates')
        >>> print(icrs)
        ICRS([[100.,  20.],
              [200.,  10.],
              [300.,   5.]])
        >>> print(icrs.pmra)
        [1, 2, 0]
        >>> print(icrs.parallax)
        [0, 0, 0]
        >>> print(icrs.name)
        my_coordinates

    Values in ``__extra_params__`` are propagated to the new instance as
    attributes and can be defined with any user-provided value.
    ``__extra_arrays__`` on the other hand are expected to be arrays with
    the same length as the number of coordinates. If the extra array value
    is not provided when instantiating the coordinate, it is replaced with
    arrays of zeros of size N (the number of coordinates). If you don't want
    the default value to be zeros you'll need to override the
    ``__array_finalize__`` method, for example ::

        def __array_finalize__(self, obj):

            super().__array_finalize__(obj)

            if getattr(obj, 'epoch', None) is None:
                self.epoch += 2451545.0

    And now the default value for ``epoch`` will be 2451545.

    `.Coordinate` takes care of slicing the arrays appropriately. If we slice
    the coordinate along the first axis the resulting object will be another
    coordinate instance in which both the internal array and the extra ones
    are sliced correctly ::

        >>> new_icrs = icrs[0:2, :]
        >>> print(new_icrs)
        ICRS([[100.,  20.],
              [200.,  10.]])
        >>> print(icrs.pmra)
        [1, 2]

    But if we slice it in a way that the resulting array cannot represent the
    coordinate system anymore, a simple Numpy array will be returned, without
    any of the additional information ::

        >>> icrs[:, 0]
        array([100., 200., 300.])

    """

    __extra_arrays__ = []
    __extra_params__ = []
    __computed_arrays__ = []  # values that are computed (not passed)
    __warn_arrays__ = []  # boolean arrays to indicate warnings
    __coord_dim__ = None

    def __new__(cls: Type[T], value, **kwargs) -> T:
        if value is not None:
            value = value.copy()  # this prevents weirdness

        obj = numpy.asarray(value, dtype=numpy.float64).view(cls)

        if len(obj.shape) != 2 or obj.shape[1] < 2:
            raise CoordinateError('The input array must be NxM')

        if obj.shape[1] < 2:
            raise CoordinateError('The input array must be NxM, M>=2.')

        if obj.__coord_dim__ is not None and obj.shape[1] != obj.__coord_dim__:
            raise CoordinateError(
                'The input array must be Nx%i.' % obj.__coord_dim__
            )

        for kwarg in kwargs:
            if kwarg not in obj.__extra_arrays__ and kwarg not in obj.__extra_params__:
                raise CoordinateError(f'Invalid input argument {kwarg!r}.')

        for param in obj.__extra_params__:
            setattr(obj, param, kwargs.get(param, None))

        # initialize extra arrays arrays
        # set to zeros if not passed
        for param in obj.__extra_arrays__:
            array = kwargs.get(param, None)
            if array is None:
                # arrays default to zeros, could cause trouble
                # but its nice for things like pm, radVal, etc
                # which should probably default to zeros
                array = numpy.zeros(obj.shape[0], dtype=numpy.float64)
            elif isinstance(array, (numpy.ndarray, list, tuple)):
                array = numpy.array(array, dtype=numpy.float64)
            else:
                array = numpy.tile(array, obj.shape[0])
            setattr(obj, param, array)

            if array.shape[0] != obj.shape[0]:
                raise CoordinateError(f'{param} size does not match '
                                      'the coordinate values.')

        # initialize computed arrays to zeros
        for param in obj.__computed_arrays__:
            array = numpy.zeros(obj.shape[0], dtype=numpy.float64)
            setattr(obj, param, array)

        # initialize warning arrays to false
        for param in obj.__warn_arrays__:
            array = numpy.array([False] * obj.shape[0])
            setattr(obj, param, array)

        return obj

    def __array_finalize__(self, obj):
        # self is the new array, obj is the old array
        if obj is None or not isinstance(obj, numpy.ndarray):
            return obj

        # This is so that copies and slices of array copy the params.
        params = []
        params += self.__extra_arrays__
        params += self.__extra_params__
        params += self.__computed_arrays__
        params += self.__warn_arrays__

        for param in params:
            setattr(self, param, getattr(obj, param, None))

    def __getitem__(self, sl):
        """ When a coordinate is sliced, slice the extra arrays too,
        and carry over params, if that's the right thing to do

        Returns
        --------
        sliced : numpy.ndarray or `.Coordinate`
            Depending on how it was sliced...
        """

        # handling everything here is tricky
        # https://scipy-cookbook.readthedocs.io/items/ViewsVsCopies.html
        # is helpful
        sliced = super().__getitem__(sl)

        if (not isinstance(sliced, numpy.ndarray) or
                sliced.ndim != 2 or sliced.shape[1] != self.shape[1]):
            return sliced.view(numpy.ndarray)

        arrays2slice = []
        arrays2slice += self.__extra_arrays__
        arrays2slice += self.__computed_arrays__
        arrays2slice += self.__warn_arrays__

        for param in arrays2slice:
            if isinstance(sl, tuple):
                # index looks something like arr[2:4,:]
                setattr(sliced, param, getattr(sliced, param)[sl[0]])
            else:
                # index looks something like arr[arr<4]
                setattr(sliced, param, getattr(sliced, param)[sl])

        return sliced

    @property
    def coordSysName(self):
        return self.__class__.__name__


class Coordinate2D(Coordinate):
    """2 dimensional coordinate system
    """

    __coord_dim__ = 2


class Coordinate3D(Coordinate):
    """3 dimensional coordinate system
    """

    __coord_dim__ = 3



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-08-16
# @Filename: test_coordio.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import ctypes

import astropy.time
import pytest

from coordio import sofa
from coordio.sofa_bindings import SOFA


def test_load():

    assert sofa._name is not None
    assert 'libsofa' in sofa._name


def test_load_fails(mocker):

    mocker.patch.object(ctypes.CDLL, '__init__', side_effect=OSError)

    with pytest.raises(OSError):
        SOFA()


def test_load_fails_with_name():

    with pytest.raises(OSError):
        SOFA('libblah.so')


def test_get_internal_date():

    assert sofa.get_internal_date() is not None

    j2000 = astropy.time.Time(2000.0, format='jyear')
    assert sofa.get_internal_date(j2000.to_datetime()) == (2451544.5, 0.5)

    now = astropy.time.Time.now()
    assert sum(sofa.get_internal_date()) == pytest.approx(now.jd, abs=1e-8)


def test_get_internal_date_error(mocker):

    mocker.patch.object(sofa, 'iauDtf2d', return_value=-1)

    with pytest.raises(ValueError):
        sofa.get_internal_date()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-08-21
# @Filename: test_time.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import astropy.time
import pytest

from coordio import IERS, Time


def test_time_now(mock_iers):

    time = Time()

    assert time.jd == pytest.approx(astropy.time.Time.now().tai.jd,
                                    abs=5e-8)

    assert time.jd1 + time.jd2 == time.jd
    assert isinstance(time.iers, IERS)
    assert time.jd - 2400000.5 == time.mjd


def test_to_scales():

    # Jan 15th, 2020
    time = Time(2458863.5, scale='TAI')
    test_time = astropy.time.Time(2458863.5, format='jd', scale='tai')

    assert time.to_utc() == pytest.approx(test_time.utc.jd, abs=1e-9)
    assert time.to_ut1() == pytest.approx(test_time.ut1.jd, abs=1e-9)
    assert time.to_tdb() == pytest.approx(test_time.tdb.jd, abs=1e-9)


def test_get_dut1():

    # Jan 15th, 2020
    time = Time(2458863.5, scale='TAI')

    assert time.get_dut1() == pytest.approx(-0.1799645, abs=1e-9)


def test_to_now():

    # Jan 15th, 2020
    time = Time(2458863.5, scale='TAI')

    time.to_now()
    assert time.jd == pytest.approx(astropy.time.Time.now().tai.jd, abs=1e-8)

    # Now test __internal_time
    time.to_now()
    assert time.jd == pytest.approx(astropy.time.Time.now().tai.jd, abs=1e-8)



#!/usr/bin/env python

"""Fit plug plate measurements

History:
2010-06-17 ROwen    2.0: first version using Python instead of Igor
2011-01-13 ROwen    2.1: Show another digit of radial position error.
                    Display radial position error as Pos Err instead of Rad Err.
                    Fit and show quadrupole moment as additional data.
2011-01-14 ROwen    2.2: Graph residuals after removing quadrupole error.
                    Normalize reported quadrupole magnitude (always positive) and angle (range -180 to 180).
                    Fixed name dipole -> quadrupole.
                    Slightly improved quadrupole fitting by using an initial value for quadrupole magnitude.
                    Improve display of long filenames (which are sometimes used for debugging).
2011-02-25 ROwen    Split out of fitPlugPlateMeas.py
                    Changed from functions to classes.
"""
import math
import numpy
import scipy.optimize

RadPerDeg = math.pi / 180.0

class Model(object):
    """A function with coefficients (of type float) that can be fit
    
    Subclasses must override basicApply
    and may wish to override setCoeffs to normalize the values
    and/or accessors to get more convenient versions of the coefficients
    """
    def __init__(self, coeffs):
        """Create a model
        
        Inputs:
        - coeffs: an array of floats (or something that can be converted to one); a copy is made
        """
        self._coeffs = numpy.array(coeffs, dtype=float)
    
    def basicApply(self, coeffs, posArr):
        """Apply the model using the specified coefficients. Subclasses must override.
        
        Inputs:
        - coeffs: an array of coefficients
        - posArr: an array of shape [n,2] of x,y positions
        
        Returns:
        - outPosArr: an array of positions after applying the model
        """
        raise RuntimeError("Subclass must override")
    
    def apply(self, posArr, doInverse=False):
        """Apply the function using the current coefficients to an array of shape [n,2]
        
        Inputs:
        - posArr: an array of shape [n,2] of x,y positions
        
        Returns:
        - adjPosArr: an array of shape [n,2] of x,y positions after applying the model
        """
        if doInverse:
            coeffs = self.getInverseCoeffs()
        else:
            coeffs = self._coeffs
        return self.basicApply(coeffs, posArr)
    
    def applyOne(self, xyPos, doInverse=False):
        """Apply the function using the current coefficients to a single point: a sequence of two floats

        Inputs:
        - xyPos: an x,y position (pair of floats)
        
        Returns:
        - adjXYPos: the x,y position after applying the model
        """
        return self.apply(numpy.array([xyPos], dtype=float), doInverse=doInverse)[0]
    
    def getCoeffs(self):
        """Get a copy of the coefficients
        """
        return numpy.array(self._coeffs, copy=True)
    
    def getInverseCoeffs(self):
        """Get coefficients for inverse transformation
        """
        raise RuntimeError("Subclass must override")
    
    def setCoeffs(self, coeffs):
        """Set the coefficients; subclasses may override to normalize the coefficients
        
        Inputs:
        - coeffs: an array of coefficients; a copy is made
        """
        if len(coeffs) != len(self._coeffs):
            raise RuntimeError("coeffs array has the wrong length")
        self._coeffs = numpy.array(coeffs, dtype=float)


class ModelFit(object):
    """Fit a Model
    """
    def __init__(self, model, measPos, nomPos, doRaise=False, maxFuncEval=5000):
        self.model = model
        self.measPos = measPos
        self.nomPos = nomPos
        
        initialCoeffs = model.getCoeffs()
        fitCoeffs, status = scipy.optimize.leastsq(
            self._computeRadSqError,
            initialCoeffs,
            maxfev = maxFuncEval,
        )
        if status not in list(range(5)):
            if doRaise:
                raise RuntimeError("fit failed")
            else:
                fitCoeffs[:] = numpy.nan
        self.model.setCoeffs(fitCoeffs)

    def _computeRadSqError(self, coeffs):
        """Compute the radial error squared for a particular set of coefficients
        
        Inputs:
        - coeffs: see applyTransRotScale
        - measPos: array of measured x,y positions
        - nomPos: array of nominal x,y positions
        """
        fitPos = self.model.basicApply(coeffs, self.nomPos)
        residPosErr = self.measPos - fitPos
        return residPosErr[:,0]**2 + residPosErr[:,1]**2
    
    def getFitPos(self):
        """Return the fit positions: model.apply(self.nomPos)
        """
        return self.model.apply(self.nomPos)
    
    def getPosError(self):
        """Return the residual position error: measPos - fitPos
        """
        return self.measPos - self.getFitPos()
    
    def getRadError(self):
        """Return the residual radial errors: radius of measPos - fitPos
        """
        return numpy.sqrt(self._computeRadSqError(self.model.getCoeffs()))


class TransRotScaleModel(Model):
    """Model translation, rotation and scale
    
    out x  =  c0  +  in x * (c3  -c2)
        y     c1        y   (c2   c3)
            thus:
            translation = (c0, c1)
            rotation = atan2(c2, c3)
            scale = sqrt(c2^2 + c3^2)
    """
    def __init__(self, coeffs=(0.0, 0.0, 0.0, 1.0)):
        """Create a model.
        
        The default coefficients are for no translation, rotation or scale change
        """
        Model.__init__(self, coeffs)

    def basicApply(self, coeffs, posArr):
        """Compute the fit position given a set of coefficients
        """
        # I want rotVec0, rotVec1,... = rotMat * [vec1, vec2,...]
        # but I don't know how to do it, so use this instead
        # rotVec0, rotVec1,... = [vec1, vec2,...] * rotMatTransposed
        rotMatTransposed = numpy.array(((coeffs[3], coeffs[2]), (-coeffs[2], coeffs[3])), dtype=float)
        return coeffs[0:2] + numpy.dot(posArr, rotMatTransposed)
    
    def getInverseCoeffs(self):
        """Get coefficients for inverse transformation
        """
        trans = (self._coeffs[0], self._coeffs[1])
        magSq = self._coeffs[2]**2 + self._coeffs[3]**2
        invC2 = -self._coeffs[2] / magSq
        invC3 = self._coeffs[3] / magSq
        invMatTransposed = numpy.array(((invC3, invC2), (-invC2, invC3)), dtype=float)
        negInvTrans = numpy.dot(trans, invMatTransposed)
        return (
            -negInvTrans[0],
            -negInvTrans[1],
            invC2,
            invC3,
        )

    def getTransRotScale(self):
        """Return translation, rotation and scale factor
        
        Returns:
        - x, y translation
        - rotation angle, in degrees
        - scale factor
        """
        return (
            self._coeffs[0:2],
            math.atan2(self._coeffs[2], self._coeffs[3]) / RadPerDeg,
            math.sqrt(self._coeffs[2]**2 + self._coeffs[3]**2),
        )
    
    def setTransRotScale(self, xyTrans, rotAng, scale):
        """Set coefficients from translation, rotation and scale factor
        """
        rotAngRad = rotAng * RadPerDeg
        self._coeffs = (
            xyTrans[0],
            xyTrans[1],
            scale * math.sin(rotAngRad),
            scale * math.cos(rotAngRad),
        )


class QuadrupoleModel(Model):
    """Model quadrupole
    
    outXY = inXY [1 + qpMag cos(2 inAng - 2 qpAng)]
            where:
            * qpMag is the quadruple magnitude
            * qpAng is the quadruple angle
          = inXY [1 + (c0 * 1e-3 * cos(2 * (inAng - c1))))
            where:
            * c0 = qpMag * 1e3 (the 1e3 helps the fitter converge)
            * apAng = c1

    WARNING: the inverse transformation is approximated as:
    outXY = inXY [1 - apMag cos(2 inAng - 2 qpAng)]
    which only works for small apMag.
    """
    def __init__(self, coeffs=(0.0, 0.0)):
        """Create a model.
        
        The default coefficients are for no quadrupole
        """
        Model.__init__(self, coeffs)

    def basicApply(self, coeffs, inPos):
        """Apply quadruple correction
        """
        xArr = inPos[:,0]
        yArr = inPos[:,1]
        rSqArr = xArr**2 + yArr**2
        inAng = numpy.arctan2(yArr, xArr)
        qpCorr = 1.0 + (coeffs[0] * 1.0e-3 * numpy.cos(2.0 * (inAng - coeffs[1])))
        qpCorr = numpy.where(rSqArr > 0.0, qpCorr, 1.0)
        return inPos * qpCorr[:,numpy.newaxis]

    def getInverseCoeffs(self):
        """Get coefficients for inverse transformation
        
        Warning: this is an approximation that only works for small quadrupole moments
        """
        # preserve magnitude and rotate angle by pi/2
        return (
            self._coeffs[0],
            self._wrapAngleCoeff(self._coeffs[1] + (math.pi * 0.5)),
        )

    def setCoeffs(self, coeffs):
        """Set coefficients
        """
        self._coeffs = numpy.array(coeffs, dtype=float)
        # make quadrupole magnitude positive, adjusting angle if necessary
        if self._coeffs[0] < 0:
            self._coeffs[0] = -self._coeffs[0]
            self._coeffs[1] += math.pi * 0.5
    
        # wrap angle coefficient
        self._coeffs[1] = self._wrapAngleCoeff(self._coeffs[1])

    def getMagnitudeAngle(self):
        """Get the quadrupole magnitude and angle
        
        Returns:
        - Quadrupole magnitude = c0
        - Quadrupole angle, in degrees = c1 converted from radians to degrees
        """
        return (
            self._coeffs[0] * 1.0e-3,
            self._coeffs[1] / RadPerDeg,
        )
    
    def setMagnitudeAngle(self, mag, ang):
        """Set coeffs from magnitude and angle

        Inputs:
        - mag: quadrupole magnitude
        - ang: quadrupole angle, in degrees
        """
        self._coeffs = (
            mag * 1.0e3,
            ang * RadPerDeg,
        )
    
    def _wrapAngleCoeff(self, angCoeff):
        """Wrap angle coeff into range [-pi/2, pi/2] (since it's degenerate to 1/2 rotation)
        
        Input:
        - angleCoeff: angle coefficient (radians)
        
        Returns:
        - wrapped angle coefficient (radians)
        """
        angCoeff = angCoeff % (2.0 * math.pi) # in range [0, 2 pi]
        if angCoeff > math.pi:
            # remove degeneracy and put in range [0, pi]
            angCoeff -= math.pi
        if angCoeff > math.pi * 0.5:
            # shift range to [-pi/2, pi/2]
            angCoeff -= math.pi
        return angCoeff


def arrayRMS(arr):
    """Return the RMS of an array
    """
    return math.sqrt(numpy.sum(arr**2) / len(arr))


if __name__ == "__main__":
    print("Test TransRotScaleModel")
    trsModel = TransRotScaleModel()
    xyPosList = []
    for xPos in (0.0, 1.0): # -1.5, 0.0, 0.5):
        for yPos in (0.0, 1.0): # (-0.5, 0.0, 1.5):
            xyPosList.append((xPos, yPos))
    xyPosArr = numpy.array(xyPosList, dtype=float)
    
    trans = (-0.5, 0.3)
    rot = 45.0
    scale = 1.2
    sqrt2 = math.sqrt(2.0)
    basicOut = numpy.array((sqrt2, sqrt2)) * scale
    trsModel.setTransRotScale(trans, rot, scale)
    for inPos, desOutPos in (
        ((1.0, 0.0), numpy.array((1.0/sqrt2, 1.0/sqrt2)) * scale + trans),
        ((0.0, 1.0), numpy.array((-1.0/sqrt2, 1.0/sqrt2)) * scale + trans),
        ((-2.0, -2.0), numpy.array((0, -sqrt2)) * 2.0 * scale + trans),
    ):
        outPos = trsModel.applyOne(inPos)
        if not numpy.allclose(outPos, desOutPos):
            print("Apply failed on coeffs=%s, inPos=%s, desOutPos=%s outPos=%s" % (trsModel.getCoeffs(), inPos, desOutPos, outPos))

    for trans in ((0.0, 0.0), (-0.5, 0.3), (0.0, 0.0), (11.9, -35.4)):
        for rot in (90.0, -180.0, -179.0, -90.0, -45.0, 0.0, 45.0, 180.0):
            for scale in (1.0, -0.1, 0.9, 0.99999, 1.0, 1.0001, 1.1, 10.0):
                trsModel.setTransRotScale(trans, rot, scale)
                cnvPosArr = trsModel.apply(xyPosArr)
                invPosArr = trsModel.apply(cnvPosArr, doInverse=True)
                if not numpy.allclose(xyPosArr, invPosArr):
                    print("Round trip failed for trans=%s, rot=%s, scale=%s" % (trans, rot, scale))
                    print("xyPosArr=", xyPosArr)
                    print("cnvPosArr=", cnvPosArr)
                    print("invPosArr=", invPosArr)

    print("Test QuadrupoleModel")
    qpModel = QuadrupoleModel()
    xyPosList = []
    for xPos in (0.0, 1.0): # -1.5, 0.0, 0.5):
        for yPos in (0.0, 1.0): # (-0.5, 0.0, 1.5):
            xyPosList.append((xPos, yPos))
    xyPosArr = numpy.array(xyPosList, dtype=float)

    # to do: add direct test of values
    
    for qpMag in (0.0, 0.001, -0.001):
        for qpAngle in (90.0, -180.0, -179.0, -90.0, -45.0, 0.0, 45.0, 180.0):
            qpModel.setMagnitudeAngle(qpMag, qpAngle)
            cnvPosArr = qpModel.apply(xyPosArr)
            invPosArr = qpModel.apply(cnvPosArr, doInverse=True)
            if not numpy.allclose(xyPosArr, invPosArr):
                print("Round trip failed for qpMag=%s, qpAngle=%s" % (qpMag, qpAngle))
                print("xyPosArr=", xyPosArr)
                print("cnvPosArr=", cnvPosArr)
                print("invPosArr=", invPosArr)



import numpy
import pytest

from coordio import Field, FocalPlane, Observed, Site, Tangent, Wok
from coordio.defaults import APO_MAX_FIELD_R, LCO_MAX_FIELD_R, calibration


def test_fpScaleAPO():
    nCoords = 200
    apoSite = Site("APO")
    apoSite.set_time(2458863, scale='TAI')
    fcAPO = Observed([[80, 120]], site=apoSite)
    phiFieldAPO = numpy.random.uniform(0, APO_MAX_FIELD_R, size=nCoords)
    thetaField = numpy.random.uniform(0, 360, size=nCoords)

    fieldAPO = Field(
        numpy.array([thetaField, phiFieldAPO]).T,
        field_center=fcAPO
    )

    focalAPOInit = FocalPlane(fieldAPO, site=apoSite, fpScale=1)

    wokRadii = []
    for sf in [0.999882, 1, 1.000118]:
        focalAPO = FocalPlane(fieldAPO, site=apoSite, fpScale=sf)
        wokAPO = Wok(focalAPO, site=apoSite, obsAngle=0)
        rad = numpy.linalg.norm(wokAPO[:,:2], axis=1)
        wokRadii.append(rad)
        _focalAPO = FocalPlane(wokAPO, site=apoSite, fpScale=sf)

        # dFocalFocalAPO = focalAPO - focalAPOInit
        # assert numpy.max(numpy.linalg.norm(dFocalFocalAPO, axis=1)) < 1e-10

        # dFocalWokAPO = focalAPO - _focalAPO
        # assert numpy.max(numpy.linalg.norm(dFocalWokAPO, axis=1)) < 1e-10

    wokRadii = numpy.array(wokRadii)
    dwokRadii = numpy.diff(wokRadii, axis=0)
    assert numpy.max(dwokRadii) < 0

if __name__ == "__main__":
    test_fpScaleAPO()


import matplotlib.pyplot as plt
import numpy
import pytest
from numpy.testing import assert_array_almost_equal

from coordio import (CoordIOUserWarning, PositionerApogee, PositionerBoss,
                     PositionerMetrology, Site, Tangent, defaults)


numpy.random.seed(0)

apoSite = Site("APO")
lcoSite = Site("LCO")

# # set time (not used but required for Observed coords)
apoSite.set_time(2458863, scale='TAI')
lcoSite.set_time(2458863, scale='TAI')

modelMetXY = numpy.array([14.314, 0])
# boss xy position in solid model
modelBossXY = numpy.array([14.965, -0.376])
# apogee xy position in solid model
modelApXY = numpy.array([14.965, 0.376])
alphaLen = 7.4 # mm

maxRadSci = numpy.linalg.norm(numpy.array([7.4, 0]) + modelBossXY)
maxRadMet = numpy.linalg.norm(numpy.array([7.4, 0]) + modelMetXY)

minRadSci = numpy.linalg.norm(numpy.array([7.4, 0]) - modelBossXY)
minRadMet = numpy.linalg.norm(numpy.array([7.4, 0]) - modelMetXY)

nCoords = 10000


def sampleAnnulus(nCoords, R1, R2):
    # R1 is max radius, R2 in min radius
    theta = 360 * numpy.random.random(size=nCoords)
    r = numpy.sqrt(numpy.random.random(size=nCoords)*(R1**2-R2**2)+R2**2)
    x = r * numpy.cos(numpy.radians(theta))
    y = r * numpy.sin(numpy.radians(theta))

    return x, y, theta, r


wl = defaults.INST_TO_WAVE["Boss"]


def test_good_coords():
    xMet, yMet, thetaMet, distMet = sampleAnnulus(
        nCoords, 0.999 * maxRadMet, 1.001 * minRadMet
    )
    xSci, ySci, thetaSci, distSci = sampleAnnulus(
        nCoords, 0.999 * maxRadSci, 1.001 * minRadSci
    )

    metXYZ = numpy.zeros((nCoords, 3))
    metXYZ[:, 0] = xMet
    metXYZ[:, 1] = yMet

    sciXYZ = numpy.zeros((nCoords, 3))
    sciXYZ[:, 0] = xSci
    sciXYZ[:, 1] = ySci

    plist = [PositionerApogee, PositionerBoss, PositionerMetrology]
    tlist = [sciXYZ, sciXYZ, metXYZ]

    for site in [apoSite, lcoSite]:
        for holeID in ["R-13C1", "R0C15"]:
            for _Pos, tanXYZ in zip(plist, tlist):
                tc = Tangent(tanXYZ, site=site, holeID=holeID, wavelength=wl)
                pc = _Pos(tc, site=site, holeID=holeID)
                assert not numpy.any(pc[:, 0] > 360)
                assert not numpy.any(pc[:, 0] < 0)
                assert not numpy.any(pc[:, 1] > 180)

                # TODO: this fails for now.
                # assert not numpy.any(pc[:, 1] < 0)

                _tc = Tangent(pc, site=site, holeID=holeID, wavelength=wl)

                # TODO: Some coordinates are nan. Until we fix it, ignore them.
                nans = numpy.any(numpy.isnan(pc), axis=1)
                tc = tc[~nans]
                pc = pc[~nans]
                _tc = _tc[~nans]

                assert True not in pc.positioner_warn
                assert numpy.isfinite(numpy.sum(pc))

                assert_array_almost_equal(tc, _tc, decimal=10)


def test_bad_coords():
    xMet, yMet, thetaMet, distMet = sampleAnnulus(nCoords, 0.9*minRadMet, 0)
    xSci, ySci, thetaSci, distSci = sampleAnnulus(nCoords, 0.9*minRadSci, 0)


    metXYZ = numpy.zeros((nCoords,3))
    metXYZ[:,0] = xMet
    metXYZ[:,1] = yMet

    sciXYZ = numpy.zeros((nCoords,3))
    sciXYZ[:,0] = xSci
    sciXYZ[:,1] = ySci

    plist = [PositionerApogee, PositionerBoss, PositionerMetrology]
    tlist = [sciXYZ, sciXYZ, metXYZ]

    for site in [apoSite, lcoSite]:
        for holeID in ["R-13C1", "R0C15"]:
            for _Pos, tanXYZ in zip(plist, tlist):
                tc = Tangent(
                    tanXYZ, site=site, holeID=holeID, wavelength=wl
                )
                pc = _Pos(tc, site=site, holeID=holeID)
                assert not False in pc.positioner_warn
                assert not numpy.isfinite(numpy.sum(pc))
                with pytest.warns(CoordIOUserWarning):
                    _tc = Tangent(pc, site=site, holeID=holeID, wavelength=wl)


def test_bad_coords2():
    xMet, yMet, thetaMet, distMet = sampleAnnulus(
        nCoords, 2 * maxRadMet, 1.1 * maxRadMet
    )
    xSci, ySci, thetaSci, distSci = sampleAnnulus(
        nCoords, 2 * maxRadSci, 1.1 * maxRadSci
    )


    metXYZ = numpy.zeros((nCoords,3))
    metXYZ[:,0] = xMet
    metXYZ[:,1] = yMet

    sciXYZ = numpy.zeros((nCoords,3))
    sciXYZ[:,0] = xSci
    sciXYZ[:,1] = ySci

    plist = [PositionerApogee, PositionerBoss, PositionerMetrology]
    tlist = [sciXYZ, sciXYZ, metXYZ]

    for site in [apoSite, lcoSite]:
        for holeID in ["R-13C1", "R0C15"]:
            for _Pos, tanXYZ in zip(plist, tlist):
                tc = Tangent(
                    tanXYZ, site=site, holeID=holeID, wavelength=wl
                )
                pc = _Pos(tc, site=site, holeID=holeID)
                assert not False in pc.positioner_warn
                assert not numpy.isfinite(numpy.sum(pc))
                with pytest.warns(CoordIOUserWarning):
                    _tc = Tangent(pc, site=site, holeID=holeID, wavelength=wl)


def test_specific():
    ab = [[0,0.001]]
    holeID = "R0C15"
    posList = [PositionerApogee, PositionerBoss, PositionerMetrology]

    pcList = []
    for Pos in posList:
        pcList.append(Pos(ab, site=apoSite, holeID=holeID))

    tcList = []
    for pc in pcList:
        tc = Tangent(pc, site=apoSite, holeID=holeID, wavelength=wl)
        print("tc", tc)
        tcList.append(tc)

    _pcList = []
    for ii in range(3):
        tc = tcList[ii]
        pc = pcList[ii]
        Pos = posList[ii]
        _pc = Pos(ab, site=apoSite, holeID=holeID)
        print(pc, _pc)
        _pcList.append(_pc)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-08-21
# @Filename: test_coordinate.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import numpy
import pytest

from coordio import CoordinateError
from coordio.coordinate import Coordinate, Coordinate2D, Coordinate3D


class _TestCoordinate(Coordinate):

    __extra_arrays__ = ['array1', 'array2']
    __extra_params__ = ['param1']
    __warn_arrays__ = ['warn']
    __computed_arrays__ = ['array12']

    def __new__(cls, value, **kwargs):
        obj = super().__new__(cls, value, **kwargs)

        array12 = obj.array1 + obj.array2
        obj.array12 = array12

        # warn if sum is greater than 10
        obj.warn = obj.array12 > 10

        return obj

class _TestCoordinate2D(Coordinate2D):
    pass

class _TestCoordinate3D(Coordinate3D):
    pass

def test_coordinate():

    test_coordinate = _TestCoordinate([[1, 2], [3, 4]])

    assert numpy.array_equal(test_coordinate, [[1, 2], [3, 4]])

    assert test_coordinate.param1 is None

    assert test_coordinate.array1 is not None
    assert test_coordinate.array2 is not None

    assert numpy.array_equal(test_coordinate.array1, [0., 0.])

    test_coordinate = _TestCoordinate([[1, 2],
                                       [3, 4]], param1=1, array1=[5, 6])
    assert test_coordinate.param1 == 1
    assert numpy.array_equal(test_coordinate.array1, [5, 6])


def test_tile():

    test_coordinate = _TestCoordinate([[1, 2], [3, 4]], array1=5)
    assert numpy.array_equal(test_coordinate.array1, [5, 5])


def test_None():

    with pytest.raises(CoordinateError) as ee:
        _TestCoordinate(None)
    assert 'The input array must be NxM' in str(ee)


def test_bad_input():

    with pytest.raises(CoordinateError) as ee:
        _TestCoordinate([[1], [2]])
    assert 'The input array must be NxM' in str(ee)

    with pytest.raises(CoordinateError) as ee:
        _TestCoordinate([[1, 2], [3, 4]], array1=numpy.array([1]))
    assert 'array1 size does not match the coordinate values.' in str(ee)

    with pytest.raises(CoordinateError):
        _TestCoordinate2D([[1,2,3],[4,5,6]])

    with pytest.raises(CoordinateError):
        _TestCoordinate3D([[1,2],[4,5]])

def test_slice():

    test_coordinate = _TestCoordinate([[1, 2],
                                       [3, 4],
                                       [5, 6]], array1=[7, 8, 9])

    sliced = test_coordinate[0:2, :]

    assert isinstance(sliced, _TestCoordinate)
    assert id(sliced) != id(test_coordinate)
    assert (sliced == numpy.array([[1, 2], [3, 4]])).all()
    assert (sliced.array1 == numpy.array([7, 8])).all()
    assert (sliced.array2 == numpy.array([0, 0])).all()

    # Confirm the original array is not modified
    assert (test_coordinate == numpy.array([[1, 2], [3, 4], [5, 6]])).all()
    assert (test_coordinate.array1 == numpy.array([7, 8, 9])).all()


def test_slice_to_numpy():

    test_coordinate = _TestCoordinate([[1, 2],
                                       [3, 4],
                                       [5, 6]], array1=[7, 8, 9])

    sliced = test_coordinate[:, 0]

    assert type(sliced).__name__ == 'ndarray'
    assert (sliced == numpy.array([1, 3, 5])).all()
    assert hasattr(sliced, 'array1') is False


def test_array_2d():

    test_coordinate = _TestCoordinate([[1, 2], [3, 4]],
                                      array1=[7, 8],
                                      array2=numpy.array([[10, 11],
                                                          [12, 13]]))

    assert test_coordinate.array2.ndim == 2

    sliced = test_coordinate[0:1, :]
    assert (sliced.array2 == numpy.array([[10, 11]])).all()


def test_copy():

    test_coordinate = _TestCoordinate([[1, 2],
                                       [3, 4],
                                       [5, 6]], array1=[7, 8, 9])

    test_copy = test_coordinate.copy()

    assert id(test_copy) != id(test_coordinate)
    assert isinstance(test_copy, _TestCoordinate)
    assert (test_coordinate == test_copy).all()
    assert (test_coordinate.array1 == test_copy.array1).all()


def test_more_slicing():
    arr = [1, 2, 3, 4, 5, 6]
    array1 = numpy.array(arr)
    array2 = numpy.array(arr)
    dim = 3
    param1 = "param1"

    coords = numpy.array([arr] * dim).T

    tc = _TestCoordinate(coords, array1=array1, array2=array2, param1=param1)

    with pytest.raises(IndexError):
        tc[:, dim]  # ask for a dim that doesn't exist

    with pytest.raises(IndexError):
        tc[6, :]  # ask for something off the end of the list

    tc1 = tc[2:3,:]
    numpy.testing.assert_equal(tc1, coords[2:3,:])
    numpy.testing.assert_equal(tc1.array1, array1[2:3])
    numpy.testing.assert_equal(tc1.array2, array2[2:3])
    assert tc1.param1 == tc.param1

    tc1 = tc[:,1:3]
    numpy.testing.assert_equal(tc1, coords[:,1:3])

    with pytest.raises(AttributeError):
        tc1.array1

    with pytest.raises(AttributeError):
        tc1.param1

    filtArr = tc.array1 < 4
    tc1 = tc[filtArr]
    numpy.testing.assert_equal(tc1, coords[filtArr])
    numpy.testing.assert_equal(tc1.array1, array1[filtArr])
    numpy.testing.assert_equal(tc1.array2, array2[filtArr])
    assert tc1.param1 == tc.param1

    filtArr = [0,3,5]
    tc1 = tc[filtArr]
    numpy.testing.assert_equal(tc1, coords[filtArr])
    numpy.testing.assert_equal(tc1.array1, array1[filtArr])
    numpy.testing.assert_equal(tc1.array2, array2[filtArr])
    assert tc1.param1 == tc.param1


def test_warn_arr():
    arr = [1, 2, 3, 4, 5, 6, 7, 8, 9]
    array1 = numpy.array(arr)
    array2 = numpy.array(arr)
    coords = numpy.array([array1, array1]).T

    array12 = array1 + array2
    warn = array12 > 10

    tc = _TestCoordinate(coords, array1=array1, array2=array2, param1="hi")

    slicer = tc.array1 >= 3
    _tc = tc[slicer]
    _coords = coords[slicer]
    _array1 = array1[slicer]
    _array2 = array2[slicer]
    _array12 = array12[slicer]
    _warn = warn[slicer]

    numpy.testing.assert_equal(_tc, _coords)
    numpy.testing.assert_equal(_tc.array1, _array1)
    numpy.testing.assert_equal(_tc.array2, _array2)
    numpy.testing.assert_equal(_tc.array12, _array12)
    numpy.testing.assert_equal(_tc.warn, _warn)
    assert _tc.param1 == "hi"

    _tc = tc[slicer, :]
    _coords = coords[slicer, :]
    _array1 = array1[slicer]
    _array2 = array2[slicer]
    _array12 = array12[slicer]
    _warn = warn[slicer]

    numpy.testing.assert_equal(_tc, _coords)
    numpy.testing.assert_equal(_tc.array1, _array1)
    numpy.testing.assert_equal(_tc.array2, _array2)
    numpy.testing.assert_equal(_tc.array12, _array12)
    numpy.testing.assert_equal(_tc.warn, _warn)
    assert _tc.param1 == "hi"

    slicer = [2,4,6]
    _tc = tc[slicer]
    _coords = coords[slicer]
    _array1 = array1[slicer]
    _array2 = array2[slicer]
    _array12 = array12[slicer]
    _warn = warn[slicer]
    assert _tc.param1 == "hi"
